/* Copyright 2003, 2004 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/

import java.io.*;
import java.net.*;
import java.nio.*;
import java.nio.channels.*;
import java.util.*;
import javax.sound.sampled.*;
import javax.sound.sampled.spi.*;

/**
 * Class that contains main method for reading a PCM signed/unsigned data
 * stream. Can be run on a file or a network stream.
 *
 * Useful URL for sound formats:
 * http://www.spies.com/~franke/SoundApp/formats.html
 */

public class ThincSound {

	// static constants
	private static final String USAGE = "Usage: ThincSound [-f <file>]|[-h <host>]";
	// 1 channel is mono, 2 for stereo, etc...
	private static final int CHANNELS = 2;
	// sample rate is how fast samples follow each other, per channel per second
	private static final float SAMPLE_RATE = 44100;
	// number of bits in the shortest sample
	private static final int SAMPLE_SIZE = 16;
	// a frame contains data for all channels at a particular time. for pcm, this is 
	// equal to the sample rate. for other encodings, the frame may contain more data.
	// usually the more compression, the slower the frame rate.
	private static final float FRAME_RATE = 44100;
	// frame size is a function of channels and sample size divided by the
	// number of bits in a byte.
	private static final int FRAME_SIZE = (CHANNELS * SAMPLE_SIZE) / 8;
	private static final AudioFormat.Encoding ENCODING = AudioFormat.Encoding.PCM_SIGNED;
	// endianness refers to how the platform swizzled the bits for the sound.
	private static final boolean IS_BIG_ENDIAN = false;

	// default port for networking
	private static final int DEFAULT_PORT = 30000;

	public static void main(String[] args) {
		if (args.length < 2 || args.length > 2) {
			System.err.println(((args.length < 2) ? "Too few " : "Too many ") 
				+ "arguments.\n" + USAGE);
			System.exit(1);
		}

		if (args[0].trim().equals("-f")) {
			playAudioFile(args[1]);			
		} else if (args[0].trim().equals("-h")) {
			playAudioNetwork(args[1]);
		} else {
			System.err.println("Unrecognized argument \"" + args[0] + "\".\n" + USAGE);
			System.exit(1);
		}

		System.exit(0);
	}

	public static void playAudioNetwork(String host) {
		SocketChannel channel;
		System.out.println("Attempting to connect to host \"" + host + "\" for playback.");
		try {
			channel = SocketChannel.open();
			channel.configureBlocking(true);
			channel.connect(new InetSocketAddress(host, DEFAULT_PORT));
			channel.finishConnect();
			ByteBuffer buf = ByteBuffer.allocate(1);
			byte b = 1;
			buf.put(b);
			channel.write(buf);
			// Now configure blocking
		} catch (IOException e) {
			System.err.println("Error when attempting to connect to host \"" + host 
				+ "\".\n" + e.toString());
			return;
		}
		System.out.println("Connected.");
		playAudio(channel);
	}

	public static void playAudioFile(String filename) {
		FileChannel channel;
		System.out.println("Attempting to open file \"" + filename + "\" for playback.");
		try {
			channel = new FileInputStream(filename).getChannel();
		} catch (FileNotFoundException e) {
			System.err.println("Could not locate file \"" + filename + "\".\n" 
				+ e.toString());
			return;
		}
		playAudio(channel);
	}

	public static void playAudio(ByteChannel channel) {

		// At present, ALAW and ULAW encodings must be converted
		// to PCM_SIGNED before it can be played
		AudioFormat	format = new AudioFormat(
					ENCODING,
					SAMPLE_RATE,
					SAMPLE_SIZE,
					CHANNELS,
					FRAME_SIZE,
					FRAME_RATE,
					IS_BIG_ENDIAN);

		// Create line
		SourceDataLine.Info info = new DataLine.Info(
			SourceDataLine.class, format);

		try {
			SourceDataLine line = (SourceDataLine) AudioSystem.getLine(info);
			line.open(format);
			line.start();
		
			// Continuously read and play chunks of audio
			int numRead = 0;
			byte[] buf = new byte[line.getBufferSize()];
			System.out.println("Buffer size: " + buf.length);
			ByteBuffer bytebuf = ByteBuffer.wrap(buf);
			bytebuf.order(ByteOrder.nativeOrder());
			int offset = 0;
			while ((numRead = channel.read(bytebuf)) > 0) {
				while (offset < numRead) {
					offset += line.write(buf, offset, numRead-offset);
				}
				bytebuf.flip();
				offset = 0;
			}
			line.drain();
			line.stop();
			channel.close();
/*
		} catch (MalformedURLException e) {
		} catch (IOException e) {
		} catch (LineUnavailableException e) {
		} catch (UnsupportedAudioFileException e) {
*/
		} catch (Exception e) {
			System.out.println("Error occurred while playing audio: " 
				+ e.toString());
		}
	}
}
