/* Copyright 2004, 2005, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Contains main function and event loop for THINC sound daemon (tsd.c).
 *
 * Authors: 
 * Leo Kim - Network Computing Lab - Columbia University
 * Rebecca Collins - Network Computing Lab - Columbia University
 * Ricardo Baratto - Network Computing Lab - Columbia University
 */
/* set tabstop = 8 */

#include <stdio.h>

/* for EXIT_FAILURE and EXIT_SUCCESS */
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <signal.h>
#include <getopt.h>

#include "tsd.h"
#include "thincProtoSound.h"

/* Local funcs */
static void tsd_exit(void);
static void tsd_exitHandler(int sig);
static void tsd_initState(void);
static int tsd_mmap(void);
static void tsd_unmap(void);
static void usage(const char *name);

/* tsd state struct */
struct tsd_state state;

#ifdef TSD_DO_TRACES
char *trace;
char *ctrace;

static void dump_trace()
{
	printf("%s\n", trace);
	ctrace = trace;
}
#endif

int main(int argc, char **argv)
{
	int port = T_SND_PORT;
	int opt, index = 0;
	struct option long_options[] = {
		{"help", 0, 0, 'h'},
		{"port", 1, 0, 'p'},
		{"no-ssl", 0, 0, 's'},
		{"udp", 0, 0, 'u'},
		{0, 0, 0, 0}	/* last row must be filled with zeroes */
	};
#ifdef TSD_DO_TRACES
	int i, *j;
#endif

	signal(SIGINT, tsd_exitHandler);

	/* initialize tsd_state */
	tsd_initState();

	/* parse options */
	while (1) {
		opt = getopt_long(argc, argv, "hp:su", long_options, &index);
		if (opt == -1) {
			/* finished */
			break;
		}

		switch (opt) {
			case 'h': 
				usage(argv[0]);
				exit(EXIT_SUCCESS);
			case 'p':
				port = (int)strtol(optarg, (char **)NULL, 10);
				if (port == EINVAL || port == ERANGE) {
					fprintf(stderr, "Invalid port "
							"specified. %s\n", 
							strerror(errno));
					usage(argv[0]);
					exit(EXIT_FAILURE);
				}
				fprintf(stdout, "Using port %d.\n", port);
				break;
			case 's':
				fprintf(stdout, "WARNING: DISABLING SSL.\n");
				state.beSecure = T_FALSE;
				break;
			case 'u':
				fprintf(stdout, "Using UDP\n");
				state.udp = T_TRUE;
				break;
			default:
				usage(argv[0]);
				exit(EXIT_FAILURE);
		}
	}

	if(state.udp) {
		state.udp_data = calloc(1, sizeof(*state.udp_data));
		if(!state.udp_data) {
			fprintf(stderr, "Out of memory: %s\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
	}

	/* open /dev/tsnd */
	if (tsd_mmap() == T_ERROR) {
		/* if this failed, then we probably didn't load our driver. */
		fprintf(stderr, "Make sure that device driver is loaded.\n");
		exit(EXIT_FAILURE);
	}

	/* now open socket */
	if (tsd_initListen(&state, port) == T_ERROR ) {
		/* if this failed, then don't bother unmapping as it means that
		 * something else is bound to the given port, probably another
		 * tsd. */
		exit(EXIT_FAILURE);
	}

	tsd_driverResetPlayCmds(state.mmapfd);

#ifdef TSD_DO_TRACES
	trace = malloc(1<<20);
	if(!trace)
		exit(EXIT_FAILURE);
	a = (int*) trace;
	for(i=0; i< (1<<16); ++i, ++a)
		*a = -1;
	ctrace = trace;

	signal(SIGHUP, dump_trace);
#endif

	while (1) {
		/* wait for something */
		if (tsd_waitForActivity(&state) == T_ERROR) {
			break;
		}
	}

	/* Wow. Did we really get here? If so, warn, then just exit. */
	fprintf(stderr, "WARNING: Somehow escaped main event loop. "
			"Attempting to close normally.\n");

	tsd_exit();

	/* Double wow. */
	fprintf(stderr, "WARNING: Could not close normally, bailing.\n");
	exit(EXIT_FAILURE);
}

/* Initialize the state structure. All new fields to tsd_state that need a 
 * default value go here. */
static void tsd_initState(void)
{
	state.clients = NULL;
	state.mmapfd = -1;
	state.sock = -1;
	state.bufsize = T_SND_BUFSIZE;
	state.ctx = NULL;
	state.beSecure = T_TRUE;
	state.playIsOpen = T_FALSE;
	state.captureIsOpen = T_FALSE;
	state.udp = T_FALSE;
	state.udp_data = NULL;

	/* initialize driver command funcs */
	tsd_initDriverCmdFuncs();
}

static int tsd_mmap(void)
{
	int fd;

	/* Open the device first */
	fprintf(stdout, "Opening %s\n", T_SND_DEVICE);
	if ( (fd = open(T_SND_DEVICE, O_RDWR)) < 0) {
		fprintf(stderr, "Couldn't open %s: %s\n", T_SND_DEVICE,
				strerror(errno));
		return T_ERROR;
	}

	/* mmap the device */
	state.play = mmap(NULL, state.bufsize, PROT_READ, MAP_PRIVATE, fd,
			0);
	state.capture = mmap(NULL, state.bufsize, PROT_READ, MAP_PRIVATE, fd,
			state.bufsize);
	if (state.play == MAP_FAILED || state.capture == MAP_FAILED) {
		fprintf(stderr, "Couldn't mmap %s: %s\n", T_SND_DEVICE, 
				strerror(errno));
		close(fd);
		return T_ERROR;
	}
	state.capture_head = state.capture;
	state.capture_tail = state.capture + state.bufsize;

	state.mmapfd = fd;

	return T_OK;
}

static void tsd_exitHandler(int sig)
{
	fprintf(stdout, "Got signal. Exiting.\n");
	tsd_exit();
}

static void tsd_unmap(void)
{
	fprintf(stdout, "Unmapping.\n");
	munmap(state.play, state.bufsize);
	munmap(state.capture, state.bufsize);
	fprintf(stdout, "Closing device.\n");
	close(state.mmapfd);
}

static void tsd_exit(void)
{
	struct tsd_client *curr, *next;
	int count = 0;

	tsd_unmap();

	/* close listening socket */
	fprintf(stdout, "Closing listening socket.\n");
	close(state.sock);

	/* disconnect all clients and free structures. */
	curr = state.clients;
	while (curr) {
		next = curr->next;
		close(curr->tcp_sock);
		close(curr->udp_sock);
		free(curr);
		curr = next;
		count++;
	}
	free(state.udp_data);

	fprintf(stdout, "Closed %d client(s).\n", count);
	fprintf(stdout, "Finished.\n");
	exit(EXIT_SUCCESS);
}

static void usage(const char *name)
{
	fprintf(stdout, "Usage: %s [options]\n", name);
	fprintf(stdout, "Options: \n"
			"\t-h/--help\t\tPrint this message and exit.\n"
			"\t-s/--no-ssl\t\tDon't use SSL for communication.\n"
			"\t-p/--port\t\tSpecify a port other than default (%d).\n"
			"\t-u/--udp\t\tUse UDP for network communication\n"
			"\t\t\tby default we use TCP\n"
			, T_SND_PORT
		   );
}
