/* Copyright 2004, 2005, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Network-related code for the THINC sound daemon.
 *
 * Authors: 
 * Leo Kim - Network Computing Lab - Columbia University
 * Rebecca Collins - Network Computing Lab - Columbia University
 * Ricardo Baratto - Network Computing Lab - Columbia University
 */
/* set tabstop = 8 */

#include <stdio.h>

/* for errno */
#include <errno.h>
#include <string.h>

#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>

#include <arpa/inet.h>
#include <netinet/in.h>

#include <sys/socket.h>

#include "tsd.h"

/* Local funcs */
static int tsd_newSocket(int port, int type);
static int tsd_setNonBlocking(int sock);
static int tsd_readSocket(struct tsd_client *client, void *buf, 
		int offset, int count);
static int tsd_writeToSocket(struct tsd_client *client, const void *buf, 
		int offset, int count);
static int tsd_readSocketSecure(struct tsd_client *client, void *buf, 
		int offset, int count);
static int tsd_writeToSocketSecure(struct tsd_client *client, const void *buf, 
		int offset, int count);
static int tsd_doPoll(int sock, short events, short *revents, int timeout);

/* even worst... */
int inline tsd_nextPort(struct tsd_state *state)
{
	static int next = 0;
	
	if(!next)
		next = state->port;

	/* FIXME: some defense against crazy long runs with many connects,
	 * but could still have problem of counter wrapping around and
	 * clashing with an existing client... oh well */
	if(next == 65534)
		next = state->port;
	return ++next;
}

/* Sets up the listening sockets. */
int tsd_initListen(struct tsd_state *state, int port)
{
	/* :HACK: for SSL support. This should be fixed throughout THINC.. */
	const char *cert_path = "/etc/X11/thinc/certs/cacert.pem";
	const char *key_path = "/etc/X11/thinc/certs/privkey.pem";

	state->sock = tsd_newSocket(port, SOCK_STREAM);
	if(state->sock == T_ERROR) {
		fprintf(stderr, "Error: Couldn't open sockets\n");
		return T_ERROR;
	}

	/* check for SSL */
	if (state->beSecure) {
		state->ctx = thinc_ssl_init_server(cert_path, key_path);
		if (!state->ctx) {
			fprintf(stderr, "Error starting tsd.\nDid you forget "
					"to put certificates in %s and %s?\n"
					"Or, did you set the permissions "
					"for these certificates to g+r,o+r?\n", 
					cert_path, key_path);
			close(state->sock);
			return T_ERROR;
		}
		fprintf(stdout, "SSL enabled.\n");
		state->read = tsd_readSocketSecure;
		state->write = tsd_writeToSocketSecure;
	} else {
		state->read = tsd_readSocket;
		state->write = tsd_writeToSocket;
	}

	state->port = port;

	return T_OK;
}

/* type is SOCK_STREAM (TCP) or SOCK_DGRAM (udp) */
static int tsd_newSocket(int port, int type)
{
	struct sockaddr_in addr_s;
	socklen_t addrlen;
	int sock, opt;

	if ( (sock = socket(PF_INET, type, 0)) == -1) {
		fprintf(stderr, "Couldn't open listening socket: %s\n",
				strerror(errno));
		return T_ERROR;
	}

	opt = 1;
	if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &opt, 
				sizeof(opt)) == -1) 
	{
		fprintf(stderr,"Error setting up listening socket: %s\n",
				strerror(errno));
		close(sock);
		return T_ERROR;
	}

	if (tsd_setNonBlocking(sock) == T_ERROR) {
		fprintf(stderr,"Error setting socket to nonblocking: %s\n",
				strerror(errno));
		close(sock);
		return T_ERROR;
	}

	addr_s.sin_family = AF_INET;
	addr_s.sin_port = htons(port);

	/* INADDR_ANY means to listen on any network interface */
	addr_s.sin_addr.s_addr = htonl(INADDR_ANY);
	addrlen = sizeof(addr_s);

	if(bind(sock, (struct sockaddr*) &addr_s, addrlen) == -1) {
		fprintf(stderr,"Error: Couldn't bind listening socket: %s.\n",
				strerror(errno));
		close(sock);
		return T_ERROR;
	}

	/* We don't anticipate much backlog for connection queues, which is 
	 * what the second argument for listen() is specifying. */
	if(type == SOCK_STREAM && ((listen(sock, 1) == -1 ))) {
		fprintf(stderr,"Error: Couldn't set up socket for "
			"listening: %s.\n", strerror(errno));
		close(sock);
		return T_ERROR;
	}

	fprintf(stdout, "Now listening to port %d\n", port);

	return sock;
}

int tsd_initClientUDP(struct tsd_client *client, struct tsd_state *state)
{
	struct sockaddr_in addr;
	socklen_t addrlen;

	client->udp_port = tsd_nextPort(state);

	client->udp_sock = tsd_newSocket(client->udp_port, SOCK_DGRAM);
	if(client->udp_sock == T_ERROR)
		return T_ERROR;

	addrlen = sizeof(addr);
	if(getpeername(client->tcp_sock, (struct sockaddr*)&addr, 
		&addrlen) == -1) 
	{
		fprintf(stderr, "Error: Couldn't get client's address: %s\n",
			strerror(errno));
		return T_ERROR;
	}

	fprintf(stdout, "Info: Got client's address %s:%d\n",
		inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));

	/* now connect to the client's UDP socket so we don't have to save
	 * this info */
	if(connect(client->udp_sock, (struct sockaddr*)&addr, addrlen) == -1) {
		fprintf(stderr, "Error: Couldn't connect to client's UDP "
			"socket: %s\n", strerror(errno));
		return T_ERROR;
	}

	return T_OK;
}

/* We need to check for three things:
 * - new client connections
 * - client disconnections
 * - sound data availability (e.g. volume, pcm)
 *
 * This function blocks indefinitely until any one of the three conditions are
 * satisfied. 
 */
int tsd_waitForActivity(struct tsd_state *state)
{
	const char *func = __func__;

	int rfdmax, res;
	struct tsd_client *client;

	/* FIXME: There's got to be a better way to keep track of these
	 * stupid sockets */
#warning "fixme: select fds sucks!"
	FD_ZERO(&state->rfds);

	/* add listening socket */
	FD_SET(state->sock, &state->rfds);
	rfdmax = state->sock;

	/* add clients */
	if (state->clients) {
		client = state->clients;
		while (client) {
			FD_SET(client->tcp_sock, &state->rfds);
			if (client->tcp_sock > rfdmax) {
				rfdmax = client->tcp_sock;
			}
			if(state->udp) {
				FD_SET(client->udp_sock, &state->rfds);
				if (client->udp_sock > rfdmax) {
					rfdmax = client->udp_sock;
				}
			}
			client = client->next;
		}
	}

	/* add driver */
	FD_SET(state->mmapfd, &state->rfds);
	if (state->mmapfd > rfdmax) {
		rfdmax = state->mmapfd;
	}

try:
	res = select(rfdmax + 1, &state->rfds, NULL, NULL, NULL);
	add_trace();
	if (res <= 0) {
		if (errno == EINTR) 
			goto try;	/* try again */
		/* what happened here... */
		fprintf(stderr, "%s: Select error. %s\n", func, 
				strerror(errno));
		return T_ERROR;
	}

	/* check for socket activity */
	tsd_handleNetworkActivity(state);

	/* check for driver activity */
	tsd_handleDriverActivity(state);

	return T_OK;
}

#if 0
int tsd_waitForActivity(struct tsd_state *state)
{
	const char *func = __func__;
	int res;
	/* All of the clients plus the mmapfd and listening socket */
	int totalPoll = state->numClients + 2;
	struct pollfd fds[totalPoll];
	struct tsd_client *client;
	state->sockPoll = state->numClients;
	state->mmapPoll = state->numClients + 1;

	/* add clients */
	if (state->clients) {
		int count = 0;
		client = state->clients;
		while (client) {
			fds[count].fd = client->sock;
			/* In the event there is data to read from the clients. */
			fds[count].events = POLLIN;
			fds[count].revents = 0;
			count++;
			client = client->next;
		}
	}

	/* add listening socket */
	fds[state->sockPoll].fd = state->sock;
	fds[state->sockPoll].events = POLLIN;
	fds[state->sockPoll].revents = 0;

	/* add driver */
	fds[state->mmapPoll].fd = state->mmapfd;
	fds[state->mmapPoll].events = POLLIN | POLLPRI;
	fds[state->mmapPoll].revents = 0;

try:
	res = poll(fds, totalPoll, -1);

	if (res == -1) {
		if (errno == EINTR) goto try;	/* try again */
		/* what happened here... */
		fprintf(stderr, "%s: Poll error. %s\n", func, strerror(errno));
		return T_ERROR;
	}

	/* okay, have the state point to the fds */
	state->fds = fds;

	/* check for driver activity */
	tsd_handleDriverActivity(state);

	/* check for socket activity */
	tsd_handleClientActivity(state);

	return T_OK;
}
#endif

int tsd_writeToClients(struct tsd_state *state, void *data, int count)
{
	const char *func = __func__;
	int written;

	struct tsd_client *client = state->clients;

	while (client) {
		written = state->write(client, data, 0, count);
		if (written < 0) {
			fprintf(stderr, "%s: Problem writing to client. %s.\n", 
					func, strerror(errno));
			return written;
		}
		client = client->next;
	}
	return T_OK;
}

/* Guarantees that count bytes are written to socket */
static int tsd_writeToSocket(struct tsd_client *client, const void *buf, 
	int offset, int count)
{
	int written = 0;
	int sock = client->tcp_sock;
	while (count > 0) {
		written = write(sock, buf + offset, count);
		if (written <= 0) {
			if (errno == EINTR) { 
				continue;
			} else if(errno == EWOULDBLOCK || errno == EAGAIN) { 
				if(tsd_doPoll(sock, POLLOUT, NULL, -1) 
						== T_ERROR)
					return T_ERROR;
				continue;
			} else {
				return T_ERROR;
			}
		}

		count -= written;
		offset += written;
	}

	return written;
}

/* SSL version of tsd_writeToSocket. Guarantees that count bytes are written */
static int tsd_writeToSocketSecure(struct tsd_client *client, const void *buf, 
		int offset, int count)
{
	int written = 0;
	SSL *ssl = client->ssl;
	int sock = client->tcp_sock;
#if 0
	fprintf(stdout, "Writing %d bytes to client.\n", count);
#endif

	while (count > 0) {
		written = thinc_ssl_write(ssl, buf + offset, count);
		if (written <= 0) {
			if (written == T_SSL_CLOSED ) {
				return T_ERROR;
			} else if (written == T_SSL_WOULDBLOCK) {
				if(tsd_doPoll(sock, POLLOUT, NULL, -1) 
						== T_ERROR)
					return T_ERROR;
				continue;
			} else {
				return T_ERROR;
			}
		}

		count -= written;
		offset += written;
	}

	return written;
}

static int tsd_doPoll(int sock, short events, short *revents, int timeout)
{
	struct pollfd fds[1];
	int res;

	fds[0].fd = sock;
	fds[0].events = events;
	fds[0].revents = 0;

again:
	res = poll(fds, 1, timeout);
	if(res == -1) {
		if(errno == EINTR)      /* try again */
			goto again;
		return T_ERROR;
	} else if(fds[0].revents & (POLLERR | POLLHUP | POLLNVAL)) {
		if(revents)
			*revents = fds[0].revents;
		return T_ERROR;
	} else if(res == 0) { /* timeout */
		return T_TIMEOUT;
	}

	if(revents)
		*revents = fds[0].revents;

	return T_OK;
}

static int tsd_setNonBlocking(int sock)
{
	int opts;

	opts = fcntl(sock,F_GETFL);
	if (opts < 0) {
		perror("fcntl(F_GETFL)");
		return T_ERROR;
	}
	opts = (opts | O_NONBLOCK);
	if (fcntl(sock,F_SETFL,opts) < 0) {
		return T_ERROR;
	}

	return T_OK;
}

/* Guarantees that count bytes are read from socket */
static int tsd_readSocket(struct tsd_client *client, void *buf, 
	int offset, int count)
{
	int nread = 0;
	int sock = client->tcp_sock;
	while (count > 0) {
		nread = read(sock, buf + offset, count);
		if (nread <= 0) {
			if (errno == EINTR) {
				continue;
			} else if(errno == EWOULDBLOCK || errno == EAGAIN) { 
				if(tsd_doPoll(sock, POLLIN, NULL, -1) 
						== T_ERROR)
					return T_ERROR;
				continue;
			} else {
				return T_ERROR;
			}
		}

		count -= nread;
		offset += nread;
	}

	return nread;
}

/* SSL version of tsd_readSocket. Guarantees that count bytes are read */
static int tsd_readSocketSecure(struct tsd_client *client, void *buf, 
		int offset, int count)
{
	int nread = 0;
	SSL *ssl = client->ssl;
	int sock = client->tcp_sock;

	while (count > 0) {
		nread = thinc_ssl_read(ssl, buf + offset, count);
		if (nread <= 0) {
			if (nread == T_SSL_CLOSED ) {
				return T_ERROR;
			} else if (nread == T_SSL_WOULDBLOCK) {
				if(tsd_doPoll(sock, POLLIN, NULL, -1) 
						== T_ERROR)
					return T_ERROR;
				continue;
			} else {
				return T_ERROR;
			}
		}

		count -= nread;
		offset += nread;
	}

	return nread;
}
