/* Copyright 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * THINC virtual sound driver. 
 *
 * Authors: 
 * Leo Kim - Network Computing Lab - Columbia University
 * Rebecca Collins - Network Computing Lab - Columbia University
 * Ricardo Baratto - Network Computing Lab - Columbia University
 */
/* set tabstop = 8 */

#include <sound/driver.h>
#include <linux/init.h>
#include <linux/jiffies.h>
#include <linux/slab.h>
#include <linux/time.h>
#include <linux/wait.h>
#include <linux/moduleparam.h>
#include <sound/core.h>
#include <sound/control.h>
#include <sound/pcm.h>
#include <sound/rawmidi.h>
#include <sound/initval.h>

#include <asm/io.h>		/* virt_to_phys() */
#include <linux/errno.h>	/* error codes */
#include <linux/fs.h>		/* everything... */
#include <linux/ioctl.h>	/* _IO* commands */
#include <linux/poll.h>		/* for poll */
#include <linux/time.h>		/* do_gettimeofday */
#include <linux/types.h>    	/* size_t */

#include <linux/version.h>	/* for KERNEL_VERSION */

#include "thincSound.h"

MODULE_AUTHOR("Leo Kim <lnk2101@cs.columbia.edu>");
MODULE_DESCRIPTION("THINC virtual sound driver");
/* TODO:
 * Uncomment this if we do GPL
 */
#if 1
MODULE_LICENSE("GPL");
#endif
MODULE_SUPPORTED_DEVICE("{{ALSA,THINC virtual sound driver}}");

#define THINC_DEBUG 0
#define THINC_DEBUG_ANNOY 0
#define THINC_STATS 0
#define ALSA_DEBUG 0
#define ALSA_DEBUG_ANNOY 0

/* :HACK: To make compilation work with kernels < 2.6.9. Otherwise, compiler
 * complains about "chip_t undeclared" errors. */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
#undef snd_kcontrol_chip
#define snd_kcontrol_chip(control) ((kcontrol)->private_data)
#endif

//#define MAX_PCM_DEVICES	4
//#define MAX_PCM_SUBSTREAMS	16
//#define MAX_MIDI_DEVICES	2

#define MAX_PCM_DEVICES		1
#define MAX_PCM_SUBSTREAMS	1
#define MAX_MIDI_DEVICES	1

/* defaults */
#ifndef MAX_BUFFER_SIZE
#if 1
/* We wish to keep this buffer size small, which apparently determines 
 * output latency. ALSA reports this buffer size to applications, which 
 * in turn use it to determine how much data an application can deliver 
 * to the driver before it starts blocking. A smaller size is important 
 * for synchronization purposes, forcing an application to write small 
 * chunks of audio data frequently and reducing latency, but too small 
 * will cause the audio to be jittery.
 */
#define MAX_BUFFER_SIZE		(16*1024)
#else
#define MAX_BUFFER_SIZE		(64*1024)
#endif
#endif
#ifndef USE_FORMATS
#define USE_FORMATS 		(SNDRV_PCM_FMTBIT_U8 | SNDRV_PCM_FMTBIT_S16_LE)
#endif
#ifndef USE_RATE
#define USE_RATE		SNDRV_PCM_RATE_CONTINUOUS | SNDRV_PCM_RATE_8000_48000
#define USE_RATE_MIN		5500
#define USE_RATE_MAX		48000
#endif
#ifndef USE_CHANNELS_MIN
#define USE_CHANNELS_MIN 	1
#endif
#ifndef USE_CHANNELS_MAX
#define USE_CHANNELS_MAX 	2
#endif
#ifndef USE_PERIODS_MIN
#define USE_PERIODS_MIN 	1
#endif
#ifndef USE_PERIODS_MAX
#define USE_PERIODS_MAX 	1024
#endif
#ifndef add_playback_constraints
#define add_playback_constraints(x) 0
#endif
#ifndef add_capture_constraints
#define add_capture_constraints(x) 0
#endif

#define T_CAPBUF_TRAIL	16384

static int index[SNDRV_CARDS] = SNDRV_DEFAULT_IDX;	/* Index 0-MAX */
static char *id[SNDRV_CARDS] = SNDRV_DEFAULT_STR;	/* ID for this card */
static int enable[SNDRV_CARDS] = {1, [1 ... (SNDRV_CARDS - 1)] = 0};
static int pcm_devs[SNDRV_CARDS] = {[0 ... (SNDRV_CARDS - 1)] = 1};
static int pcm_substreams[SNDRV_CARDS] = {[0 ... (SNDRV_CARDS - 1)] = 1};
static int midi_devs[SNDRV_CARDS] = {[0 ... (SNDRV_CARDS - 1)] = 1};

/* In 2.6 kernels <= 2.6.9, you need to pass to module_parm_array an
 * uninitialized static int instead of NULL in order for this to compile. 
 */
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,9)
static int hack;
module_param_array(index, int, hack, 0444);
module_param_array(id, charp, hack, 0444);
module_param_array(enable, bool, hack, 0444);
module_param_array(pcm_devs, int, hack, 0444);
module_param_array(pcm_substreams, int, hack, 0444);
module_param_array(midi_devs, int, hack, 0444);
#else
module_param_array(index, int, NULL, 0444);
module_param_array(id, charp, NULL, 0444);
module_param_array(enable, bool, NULL, 0444);
module_param_array(pcm_devs, int, NULL, 0444);
module_param_array(pcm_substreams, int, NULL, 0444);
module_param_array(midi_devs, int, NULL, 0444);
#endif

MODULE_PARM_DESC(index, "Index value for THINC soundcard.");
MODULE_PARM_DESC(id, "ID string for THINC soundcard.");
MODULE_PARM_DESC(enable, "Enable this THINC soundcard.");
MODULE_PARM_DESC(pcm_devs, "PCM devices # (0-4) for THINC driver.");
MODULE_PARM_DESC(pcm_substreams, "PCM substreams # (1-16) for THINC driver.");
MODULE_PARM_DESC(midi_devs, "MIDI devices # (0-2) for THINC driver.");

#define MIXER_ADDR_MASTER	0
#define MIXER_ADDR_LINE		1
#define MIXER_ADDR_MIC		2
#define MIXER_ADDR_SYNTH	3
#define MIXER_ADDR_CD		4
#define MIXER_ADDR_LAST		4

/* gettime in microseconds */
#define gettime_us(t) do { \
	struct timeval __tv;        \
	do_gettimeofday(&__tv);  \
	t = (__tv.tv_sec*1000000) + (__tv.tv_usec); \
} while(0)

/* The model for the queue here is somewhat borrowed from the kfifo
 * implementation in the 2.6.10 kernel. Nonetheless, we have some of our own
 * customizations added, and we should be compatible with at least the entire
 * 2.6.x series. 
 *
 * The way our fifo works is as follows. There are two portions to the buffer, 
 * one for holding pure audio data and the other for holding commands to tsd.
 * tsd issues an mmap call to gain access to this memory region and reads 
 * data from it. The audio data are held in the "left" side of the buffer, and 
 * the command data goes on the "right" side. In other words:
 *  
 * Audio data: 
 *	0 to T_SND_SOUND_BUFSIZE - 1
 * Command data: 
 *	T_SND_SOUND_BUFSIZE to (T_SND_SOUND_BUFSIZE + T_SND_CMD_BUFSIZE)
 * 
 * Neither audio or command data are expected to wrap around the buffer 
 * evenly, so if there is not enough space remaining in the buffer for a given
 * chunk of data, it is expected to be written at the *beginning* of the buffer.
 *
 * A command is created whenever:
 *	- Sound data are written to the buffer.
 * 	- The driver is opened.
 *	- The driver is closed.
 *
 * There are two command queues -- one is read via a read() command and the 
 * other is read via the mmap buffer. All of the playback commands are written 
 * to the mmap buffer for faster processing, and all control commands are 
 * written to a list structure. Control commands are given priority over 
 * playback commands, which is the reason for a separate queue. 
 *
 * Each time a command is created, a counter is incremented for the appropriate
 * queue. If the counter is > 0, then the driver awakens tsd (assumed to be 
 * polling on it) and tsd is expected to process this command. Once tsd processes 
 * the command, tsd decrements the counter using an ioctl. If the counter reaches 
 * 0, tsd is allowed to sleep on poll. (Though whether tsd decides to poll is 
 * entirely up to tsd.)
 *
 * IMPORTANT: No attempt is made to prevent overwriting data within the driver. 
 * The driver is expected to keep pointers to the latest hunk of data inserted 
 * in the buffer, and that's it. tsd should drop audio data if it is not 
 * transmittable at the proper time, which is what most audio experts agree is 
 * a standard practice for networked audio.
 *
 * Note: we need to make some compile-time changes to this module
 * since the MODULE_PARM calls were changed slightly in 2.6.10. 
 * (see :IMPORTANT: comment above) 
 */
struct thinc_snd_fifo {
	/* playback specific things */
	/* points to the beginning of the buffer */
	unsigned char *buf;
	/* protects concurrent modifications on buf */
	spinlock_t buf_lock;
	/* pointer to end of sound buffer */	
	unsigned int snd_in;
	/* pointer to last play command NOT PROCESSED */
	unsigned int play_last;
	/* pointer to end of play command buffer */	
	unsigned int play_in;

	/* capture specific things */
	unsigned char *capbuf;
#if 0	/* CRAP */
	int cb_length;
	int cb_numblocks;
	int cb_begin;
	int cb_end;
#else
	/* how much data we have available */
	int capbuf_avail;
	/* where to read data from on next capture */
	unsigned char *capbuf_head;
	/* tail of the capture buffer (so we can handle wraps) */
	unsigned char *capbuf_tail;
#endif

	/* number of commands currently in the play command buffer. 
	 * we don't expect too many, so we use a short */
	unsigned short num_play_cmds;
	/* number of commands currently in the control command buffer. 
	 * we don't expect too many, so we use a short */
	unsigned short num_ctrl_cmds;
	/* counter that tells us whether tsd has opened us or not. */
	unsigned short connected;
	/* required for poll() to hold blocking processes */
	wait_queue_head_t waitq_in;
	/* useful when tsd connects to driver while soudn is playing */
	struct thinc_snd_cmd_open open_params;
	/* command queue for control commands */
	struct list_head ctrl_commands;

#if THINC_STATS
	/* total number of bytes written per play instance */
	unsigned int total;
#endif
};

/*
 * kernel folks don't like typedefs in the code, so newer kernel have
 * less of them than older kernels. to easily accommodate for that, we
 * use "ts_xxx": becomes "xxx_t" in old kernel, and "struct xxx" in new
 */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,12)
#define ts_snd_pcm_t                 struct snd_pcm
#define ts_snd_card_t                struct snd_card
#define ts_snd_pcm_substream_t       struct snd_pcm_substream
#define ts_snd_pcm_runtime_t         struct snd_pcm_runtime
#define ts_snd_pcm_hardware_t        struct snd_pcm_hardware
#define ts_snd_pcm_ops_t             struct snd_pcm_ops
#define ts_snd_kcontrol_t            struct snd_kcontrol
#define ts_snd_kcontrol_new_t        struct snd_kcontrol_new
#define ts_snd_ctl_elem_value_t      struct snd_ctl_elem_value
#define ts_snd_ctl_elem_info_t       struct snd_ctl_elem_info
#else
#define ts_snd_pcm_t                 snd_pcm_t
#define ts_snd_card_t                snd_card_t
#define ts_snd_pcm_substream_t       snd_pcm_substream_t
#define ts_snd_pcm_runtime_t         snd_pcm_runtime_t
#define ts_snd_pcm_hardware_t        snd_pcm_hardware_t
#define ts_snd_pcm_ops_t             snd_pcm_ops_t
#define ts_snd_kcontrol_t            snd_kcontrol_t
#define ts_snd_kcontrol_new_t        snd_kcontrol_new_t
#define ts_snd_ctl_elem_value_t      snd_ctl_elem_value_t
#define ts_snd_ctl_elem_info_t       snd_ctl_elem_info_t
#endif


/* For command queue for tsd. This encapsulates a command, but also uses the
 * list_head node from the kernel to manage the queue of commands. */
struct thinc_snd_cmd_node {
	struct thinc_snd_cmd cmd;
	struct list_head list;
};

typedef struct snd_card_thinc {
	ts_snd_card_t *card;
	spinlock_t mixer_lock;
	int mixer_volume[MIXER_ADDR_LAST+1][2];
	int capture_source[MIXER_ADDR_LAST+1][2];
} snd_card_thinc_t;

typedef struct snd_card_thinc_pcm {
	snd_card_thinc_t *thinc;
	spinlock_t lock;
	struct timer_list timer;
	unsigned int pcm_size;
	unsigned int pcm_count;
	unsigned int pcm_bps;		/* bytes per second */
	unsigned int pcm_jiffie;	/* bytes per one jiffie */
	unsigned int pcm_irq_pos;	/* IRQ position */
	unsigned int pcm_buf_pos;	/* position in buffer */
	ts_snd_pcm_substream_t *substream;
	struct thinc_snd_fifo *fifo;
} snd_card_thinc_pcm_t;

static ts_snd_card_t *snd_thinc_cards[SNDRV_CARDS] = SNDRV_DEFAULT_PTR;

/* :TODO: For multiple user support, should have support for a list of users. */
struct thinc_snd_fifo *thinc_snd_fifo;

/***************************** file operations *********************************/
/* tsd device functions */
int thinc_snd_mmap(struct file *filp, struct vm_area_struct *vma);
int thinc_snd_open(struct inode *inode, struct file *filp);
ssize_t thinc_snd_read(struct file *filp, char *buf, size_t count,
	loff_t *offset);
ssize_t thinc_snd_write(struct file *filp, const char *buf, size_t count, 
	loff_t *offset);
int thinc_snd_release(struct inode *inode, struct file *filp);
int thinc_snd_ioctl(struct inode *inode, struct file *filp, unsigned int cmd,
	unsigned long arg);
unsigned int thinc_snd_poll(struct file *filp, poll_table *wait);

/* audio playback functions */
static int snd_card_thinc_playback_silence(ts_snd_pcm_substream_t *substream,
	int channel, snd_pcm_uframes_t pos, snd_pcm_uframes_t count);
static int snd_card_thinc_playback_copy(ts_snd_pcm_substream_t *substream, 
	int channel, snd_pcm_uframes_t pos, void __user *src, 
	snd_pcm_uframes_t count);
/* audio capture functions */
static int snd_card_thinc_capture_silence(ts_snd_pcm_substream_t *substream, 
	int channel, snd_pcm_uframes_t pos, snd_pcm_uframes_t count);
static int snd_card_thinc_capture_copy(ts_snd_pcm_substream_t *substream, 
	int channel, snd_pcm_uframes_t pos, void __user *src, 
	snd_pcm_uframes_t count);

#if 0
static int snd_card_thinc_playback_copy_noop(ts_snd_pcm_substream_t *substream,
	int channel, snd_pcm_uframes_t pos, void __user *src, 
	snd_pcm_uframes_t count);
static int snd_card_thinc_playback_prepare_noop(ts_snd_pcm_substream_t *substream);
#endif

/* fifo functions */
static inline int fifo_get_sound(struct thinc_snd_fifo *fifo, 
	unsigned char __user *user_buf, unsigned int len);
static inline int fifo_put_sound(struct thinc_snd_fifo *fifo, 
	unsigned char __user *user_buf, unsigned int stamp, unsigned int len);
static int __fifo_put_sound(struct thinc_snd_fifo *fifo,
	unsigned char __user *user_buf, unsigned int stamp, unsigned int len);
static inline int __fifo_get_sound(struct thinc_snd_fifo *fifo, 
	unsigned char __user *user_buf, unsigned int len);
static struct thinc_snd_fifo *fifo_init(void);

/* command functions. see T_MODE_* for values of mode */
static int thinc_snd_cmd_open(ts_snd_pcm_runtime_t *runtime, 
	struct thinc_snd_fifo *fifo, int mode);
static int thinc_snd_cmd_close(struct thinc_snd_fifo *fifo, int mode);
static int thinc_snd_cmd_play(struct thinc_snd_fifo *fifo, 
	unsigned int offset, unsigned int stamp, unsigned int len);
#if 0
/* TODO */
static int thinc_snd_cmd_volume_down(ts_snd_pcm_runtime_t *runtime, 
	struct thinc_snd_fifo *fifo);
static int thinc_snd_cmd_volume_up(ts_snd_pcm_runtime_t *runtime,
	struct thinc_snd_fifo *fifo);
#endif
static struct thinc_snd_cmd_node *thinc_snd_make_cmd_node(
		struct thinc_snd_fifo *fifo, int id);
static void thinc_snd_free_cmds(struct thinc_snd_fifo *fifo);

/* debug functions */
#if ALSA_DEBUG
static void debug_substream(ts_snd_pcm_substream_t *substream);
#endif

/******************************* fifo functions *******************************/
static struct thinc_snd_fifo *fifo_init(void)
{
	unsigned char *play, *capture;
	struct thinc_snd_fifo *fifo;

	/* We previously used to kmalloc the pages and apply SetPageReserved on
	 * each of the pages ourselves. For some reason, this would not work.
	 * snd_malloc_pages appears to do what we want, though. */
	play = snd_malloc_pages(T_SND_BUFSIZE, GFP_KERNEL);
	if (!play) 
		return ERR_PTR(-ENOMEM);

	capture = snd_malloc_pages(T_SND_BUFSIZE, GFP_KERNEL);
	if (!capture) {
		kfree(play);
		return ERR_PTR(-ENOMEM);
	}

	fifo = kmalloc(sizeof(struct thinc_snd_fifo), GFP_KERNEL);
	if (!fifo) {
		kfree(capture);
		kfree(play);
		return ERR_PTR(-ENOMEM);
	}

	/* Zeroing out data in buffer. Do we really need to bother if the tsd 
	 * is to only obey the pointers we give it? */
	memset(play, 0, T_SND_BUFSIZE);
	fifo->buf = play;
	fifo->num_ctrl_cmds = fifo->num_play_cmds = 0;
	fifo->snd_in = T_SND_SOUND_BUFSTART;
	fifo->play_last = fifo->play_in = T_SND_CMD_BUFSTART;

	/* audio capture init */
	memset(capture, 0, T_SND_BUFSIZE);
	fifo->capbuf = capture;
#if 0	/* CRAP */
	fifo->cb_length = T_SND_BUFSIZE / T_SND_CAPTURE_BLOCKSIZE;
	fifo->cb_begin = fifo->cb_end = 0;
	fifo->cb_numblocks = 0;
#else
	fifo->capbuf_avail = -1;
	fifo->capbuf_head = fifo->capbuf;
	fifo->capbuf_tail = fifo->capbuf + T_SND_BUFSIZE;
#if THINC_DEBUG
	printk(KERN_INFO "thinc_snd: init, have %d bytes available\n", 
		fifo->capbuf_avail);
#endif	
#endif

	/* Init locks */
	spin_lock_init(&fifo->buf_lock);

	/* Init ctrl command list */
	INIT_LIST_HEAD(&fifo->ctrl_commands);

	init_waitqueue_head(&fifo->waitq_in);
	fifo->connected = 0;

	return fifo;
}

/* Resets the number of playback commands to process to zero. */
static inline void fifo_reset_play_cmds(struct thinc_snd_fifo *fifo)
{           
	unsigned long flags;
	spin_lock_irqsave(&fifo->buf_lock, flags);
	fifo->num_play_cmds = 0;
	fifo->play_last = fifo->play_in;
	spin_unlock_irqrestore(&fifo->buf_lock, flags);
}

/* Resets the number of playback commands to process to zero. */
static inline void fifo_del_play_cmds(struct thinc_snd_fifo *fifo)
{           
	unsigned long flags;
	unsigned int new;
	spin_lock_irqsave(&fifo->buf_lock, flags);
	fifo->num_play_cmds--;
	new = fifo->play_last + sizeof(struct thinc_snd_cmd);
	if (new + sizeof(struct thinc_snd_cmd) > T_SND_CMD_BUFEND) {
		new = T_SND_CMD_BUFSTART;
	}
	fifo->play_last = new;
	spin_unlock_irqrestore(&fifo->buf_lock, flags);
}

/* This is its own separate function since its operations rely on being in its
 * own space in the call stack. assumes len is kosher and fifo buf lock is
 * taken */
static int __fifo_put_sound(struct thinc_snd_fifo *fifo, 
	unsigned char __user *user_buf, unsigned int stamp, unsigned int len)
{
	unsigned int new;

	new = fifo->snd_in + len;
	if (new > T_SND_SOUND_BUFEND) {
		fifo->snd_in = T_SND_SOUND_BUFSTART;
		new = fifo->snd_in + len;
	}

	if (copy_from_user(fifo->buf + fifo->snd_in, user_buf, len)) {
		return -EFAULT;
	}

	/* generate play command, which wakes things up. */
	thinc_snd_cmd_play(fifo, fifo->snd_in, stamp, len);

	fifo->snd_in = new;

	/* wake up processes in waitqueue to process command */
	wake_up(&fifo->waitq_in);

	return 0;
}

/* This is its own separate function since its operations rely on being in its
 * own space in the call stack. assumes len is kosher and locks are taken */
#if 0
static inline int __fifo_get_sound(struct thinc_snd_fifo *fifo, 
	unsigned char __user *user_buf, unsigned int len)
{
	int to_end;

	/* check if available data wraps around the buffer */
	to_end = fifo->capbuf_tail - fifo->capbuf_head;
	if(len > to_end) {
		/* it does wrap around :( */
		if (copy_to_user(user_buf, fifo->capbuf_head, to_end)) {
			return -EFAULT;
		}
		if (copy_to_user(user_buf + to_end, fifo->capbuf, len-to_end)) {
			return -EFAULT;
		}
		fifo->capbuf_head = fifo->capbuf + len - to_end;
	} else {
		/* all contiguous */
		if (copy_to_user(user_buf, fifo->capbuf_head, len)) {
			return -EFAULT;
		}
		fifo->capbuf_head += len;
	}
	fifo->capbuf_avail -= len;
#if THINC_DEBUG
	printk(KERN_INFO "thinc_snd: copied %d bytes of sound, have %d left\n", 
		len, fifo->capbuf_avail);
#endif
	if(fifo->capbuf_head == fifo->capbuf_tail)
		fifo->capbuf_head = fifo->capbuf;

	return 0;
}
#else
/* 
 * assumes you checked there's enough data.
 * the way this works is kind of funky so pay attention.
 * - once we open the device for capture, the client starts reading data
 *   and sending it to us
 * - tsd reads this data, puts it in our buffer, and tells us how much data
 *   it wrote. we then update capbuf_avail.
 * - when app asks for data, this function gets called to copy the data
 *   back to userland. 
 * - now if we do the stupid thing and just return from the head of our
 *   buffer, we get a huge delay (however long between when the client 
 *   started pushing data and the application actually started reading)
 * - so what we do is return len bytes counted from the tail
 *   of the buffer, and just discard whatever data is before.
 */
static inline int __fifo_get_sound(struct thinc_snd_fifo *fifo, 
	unsigned char __user *user_buf, unsigned int len)
{
	int from_head;
	char *from;

	/* we may destroy len later on */
	fifo->capbuf_avail -= len;

	/* check if available data wraps around the buffer */
	from_head = fifo->capbuf_head - fifo->capbuf;
	if(from_head < len) {
		/* wrap around :( this is kind of fun */
		len -= from_head;
		from = fifo->capbuf_tail - len;
		if(copy_to_user(user_buf, from, len)) {
			return -EFAULT;
		}
		if(copy_to_user(user_buf + len, fifo->capbuf, from_head)) {
			return -EFAULT;
		}
	} else {
		/* no wrap around, wee */
		from = fifo->capbuf_head - len;
		if(copy_to_user(user_buf, from, len)) {
			return -EFAULT;
		}
	}
#if THINC_DEBUG
	printk(KERN_INFO "thinc_snd: copied %d bytes of sound, have %d left\n", 
		len, fifo->capbuf_avail);
#endif
	return 0;
}
#endif

static inline int fifo_put_sound(struct thinc_snd_fifo *fifo, 
	unsigned char __user *user_buf, unsigned int stamp, unsigned int len)
{   
	unsigned long flags;
	int ret;

	spin_lock_irqsave(&fifo->buf_lock, flags);
	ret = __fifo_put_sound(fifo, user_buf, stamp, len);
	spin_unlock_irqrestore(&fifo->buf_lock, flags);

	return ret;
}

static inline int fifo_get_sound(struct thinc_snd_fifo *fifo, 
	unsigned char __user *user_buf, unsigned int len)
{   
	unsigned long flags;
	int ret;

	spin_lock_irqsave(&fifo->buf_lock, flags);
	ret = __fifo_get_sound(fifo, user_buf, len);
	spin_unlock_irqrestore(&fifo->buf_lock, flags);

	return ret;

}

/***************************** command functions ******************************/
static int thinc_snd_cmd_open(ts_snd_pcm_runtime_t *runtime, 
	struct thinc_snd_fifo *fifo, int mode)
{
	struct thinc_snd_cmd_node *node;

#if THINC_DEBUG
	printk(KERN_INFO "thinc_snd_cmd_open: Creating open command\n");
#endif

	if (fifo->connected == 0) {
		return 0;
	}

	/* Create initial message for tsd */
	if(mode == T_MODE_PLAYBACK) {
		node = thinc_snd_make_cmd_node(fifo, T_SND_CMD_PLAYBACK_OPEN);
	} else { /* T_MODE_CAPTURE */
		node = thinc_snd_make_cmd_node(fifo, T_SND_CMD_CAPTURE_OPEN);
	}
	if (!node) {
		return -ENOMEM;
	}

	node->cmd.type.open.bits = runtime->sample_bits;
	node->cmd.type.open.channels = runtime->channels;
	node->cmd.type.open.rate = runtime->rate;
	node->cmd.type.open.endian = 
		(snd_pcm_format_big_endian(runtime->format)) ? T_BIG : T_LITTLE;

	/* keep copy of this command in fifo struct for ioctl purposes. */
	memcpy(&fifo->open_params, &node->cmd.type.open,  
			sizeof(struct thinc_snd_cmd_open));

	/* wake up processes in waitqueue to process command */
	wake_up(&fifo->waitq_in);

	return 0;
}

static int thinc_snd_cmd_close(struct thinc_snd_fifo *fifo, int mode)
{
	struct thinc_snd_cmd_node *node;

	if (!fifo) {
#if THINC_DEBUG_ANNOY
		printk(KERN_WARNING "thinc_snd: fifo was null in "
				"thinc_snd_cmd_close!!!.\n");
#endif
		return 0;
	}
#if THINC_DEBUG_ANNOY
	printk(KERN_WARNING "thinc_snd: thinc_snd_cmd_close checking if "
			"there is anyone connected.\n");
#endif

	if (fifo->connected == 0) {
#if THINC_DEBUG_ANNOY
		printk(KERN_WARNING "thinc_snd: thinc_snd_cmd_close sez "
				"there's no one connected.\n");
#endif
		return 0;
	}

	if(mode == T_MODE_PLAYBACK)
		node = thinc_snd_make_cmd_node(fifo, T_SND_CMD_PLAYBACK_END);
	else
		node = thinc_snd_make_cmd_node(fifo, T_SND_CMD_CAPTURE_END);
	if (node == NULL) {
		return -ENOMEM;
	}

	/* wake up processes in waitqueue to process command */
	wake_up(&fifo->waitq_in);

	return 0;
}

/* this is called within fifo_put_sound. */
static int thinc_snd_cmd_play(struct thinc_snd_fifo *fifo,
		unsigned int offset, unsigned int stamp, unsigned int len)
{
	struct thinc_snd_cmd *cmd;
	unsigned int new; 

#if THINC_DEBUG
	printk(KERN_INFO "thinc_snd_cmd_play: Creating play command\n");
#endif

	new = fifo->play_in + sizeof(struct thinc_snd_cmd);

	if (new > T_SND_CMD_BUFEND) {
		fifo->play_in = T_SND_CMD_BUFSTART;
		new = fifo->play_in + sizeof(struct thinc_snd_cmd);
	}

	cmd = (struct thinc_snd_cmd *)(fifo->buf + fifo->play_in);

	cmd->id = T_SND_CMD_PLAY;
	cmd->type.play.offset = offset;
	cmd->type.play.timestamp = stamp;
	cmd->type.play.length = len;

	fifo->play_in = new;
	fifo->num_play_cmds++;

	return 0;
}


static struct thinc_snd_cmd_node *thinc_snd_make_cmd_node(
	struct thinc_snd_fifo *fifo, int id)
{   
	struct thinc_snd_cmd_node *node;

	/* Make sure that we have a fifo to actually add this command to */
	if (!fifo) {
		/* We enforce the fact that if no one is connected, then apps
		 * should not be able to playback. */
		return NULL;
	}

	node = kmalloc(sizeof(struct thinc_snd_cmd_node), GFP_KERNEL);
	if (!node) {
		printk(KERN_WARNING "snd-thinc: COULD NOT ALLOCATE MEMORY FOR "
				"COMMAND.\n");
		return NULL;
	}

	node->cmd.id = id;

	/* Insert into list */
	list_add_tail(&node->list, &fifo->ctrl_commands);

	return node; 
}       

static void thinc_snd_free_cmds(struct thinc_snd_fifo *fifo)
{   
	struct list_head *p, *n;
	struct thinc_snd_cmd_node *node;

	list_for_each_safe(p, n, &fifo->ctrl_commands) {
		node = list_entry(p, struct thinc_snd_cmd_node, list);
		list_del(&node->list);
		kfree(node);
	}
}


/****************************** driver functions ******************************/

static void snd_card_thinc_pcm_timer_start(ts_snd_pcm_substream_t * substream)
{
	ts_snd_pcm_runtime_t *runtime = substream->runtime;
	snd_card_thinc_pcm_t *dpcm = runtime->private_data;
#if ALSA_DEBUG
	printk(KERN_INFO "thinc_snd: snd_card_thinc_pcm_timer_start.\n");
#endif

	dpcm->timer.expires = 1 + jiffies;
	add_timer(&dpcm->timer);
}

static void snd_card_thinc_pcm_timer_stop(ts_snd_pcm_substream_t * substream)
{
	ts_snd_pcm_runtime_t *runtime = substream->runtime;
	snd_card_thinc_pcm_t *dpcm = runtime->private_data;
#if ALSA_DEBUG
	printk(KERN_INFO "thinc_snd: snd_card_thinc_pcm_timer_stop.\n");
#endif
	del_timer(&dpcm->timer);
}

static int snd_card_thinc_playback_trigger(ts_snd_pcm_substream_t * substream,
		int cmd)
{
	if (cmd == SNDRV_PCM_TRIGGER_START) {
#if ALSA_DEBUG
		printk(KERN_INFO "thinc_snd: snd_card_thinc_playback_trigger: "
				"START.\n");
#endif
		snd_card_thinc_pcm_timer_start(substream);
	} else if (cmd == SNDRV_PCM_TRIGGER_STOP) {
#if ALSA_DEBUG
		printk(KERN_INFO "thinc_snd: snd_card_thinc_playback_trigger: "
				"STOP.\n");
#endif
		snd_card_thinc_pcm_timer_stop(substream);
	} else {
		return -EINVAL;
	}
	return 0;
}

static int snd_card_thinc_capture_trigger(ts_snd_pcm_substream_t * substream,
		int cmd)
{
	if (cmd == SNDRV_PCM_TRIGGER_START) {
		snd_card_thinc_pcm_timer_start(substream);
	} else if (cmd == SNDRV_PCM_TRIGGER_STOP) {
		snd_card_thinc_pcm_timer_stop(substream);
	} else {
		printk(KERN_WARNING "thinc_snd: snd_card_thinc_capture_trigger:"
			" unimplemented %d\n", cmd);
		return -EINVAL;
	}
	return 0;
}

/* This function sets up the "fake" virtual buffer positioning. This buffer 
 * size is determined by MAX_BUFFER_SIZE, set in snd_card_thinc_playback and
 * snd_card_thinc_capture. */
static int snd_card_thinc_pcm_prepare(ts_snd_pcm_substream_t * substream,
	int mode)
{
	ts_snd_pcm_runtime_t *runtime = substream->runtime;
	snd_card_thinc_pcm_t *dpcm = runtime->private_data;
	unsigned int bps;

#if ALSA_DEBUG
	printk(KERN_INFO "thinc_snd: snd_card_thinc_pcm_prepare, mode %d\n", 
		mode);
	debug_substream(substream);
#endif

	bps = runtime->rate * runtime->channels;
	bps *= snd_pcm_format_width(runtime->format);
	bps /= 8;
	if (bps <= 0)
		return -EINVAL;
	dpcm->pcm_bps = bps;
	dpcm->pcm_jiffie = bps / HZ;
	dpcm->pcm_size = snd_pcm_lib_buffer_bytes(substream);
	dpcm->pcm_count = snd_pcm_lib_period_bytes(substream);
	dpcm->pcm_irq_pos = 0;
	dpcm->pcm_buf_pos = 0;

	/* :TODO: how can we have UID-specific fifos? */
	dpcm->fifo = thinc_snd_fifo;

	/* Create initial message for tsd */
	return thinc_snd_cmd_open(runtime, thinc_snd_fifo, mode);
}

static int snd_card_thinc_playback_prepare(ts_snd_pcm_substream_t * substream)
{
	return snd_card_thinc_pcm_prepare(substream, T_MODE_PLAYBACK);
}

static int snd_card_thinc_capture_prepare(ts_snd_pcm_substream_t * substream)
{
	struct thinc_snd_fifo *fifo = thinc_snd_fifo; 
	unsigned long flags;

	spin_lock_irqsave(&fifo->buf_lock, flags);
	fifo->capbuf_avail = 0;

/* XXX DEBUG */
#if 0
{
unsigned long us;
gettime_us(us);
printk(KERN_INFO "thinc_snd: [%lu] open, h: %d\n", us, fifo->capbuf_avail);
}
#endif
	spin_unlock_irqrestore(&fifo->buf_lock, flags);

	return snd_card_thinc_pcm_prepare(substream, T_MODE_CAPTURE);
}

static void snd_card_thinc_pcm_timer_function(unsigned long data)
{
	snd_card_thinc_pcm_t *dpcm = (snd_card_thinc_pcm_t *)data;

	dpcm->timer.expires = 1 + jiffies;
	add_timer(&dpcm->timer);
	spin_lock_irq(&dpcm->lock);
	dpcm->pcm_irq_pos += dpcm->pcm_jiffie;
	dpcm->pcm_buf_pos += dpcm->pcm_jiffie;
	dpcm->pcm_buf_pos %= dpcm->pcm_size;
	if (dpcm->pcm_irq_pos >= dpcm->pcm_count) {
		dpcm->pcm_irq_pos %= dpcm->pcm_count;
		snd_pcm_period_elapsed(dpcm->substream);
	}
	spin_unlock_irq(&dpcm->lock);	
}

static snd_pcm_uframes_t snd_card_thinc_playback_pointer(
	ts_snd_pcm_substream_t * substream)
{
	ts_snd_pcm_runtime_t *runtime = substream->runtime;
	snd_card_thinc_pcm_t *dpcm = runtime->private_data;

	return bytes_to_frames(runtime, dpcm->pcm_buf_pos);
}

static snd_pcm_uframes_t snd_card_thinc_capture_pointer(
	ts_snd_pcm_substream_t * substream)
{
	ts_snd_pcm_runtime_t *runtime = substream->runtime;
	snd_card_thinc_pcm_t *dpcm = runtime->private_data;

	return bytes_to_frames(runtime, dpcm->pcm_buf_pos);
}

static ts_snd_pcm_hardware_t snd_card_thinc_playback =
{

#if 0
	.info =		(SNDRV_PCM_INFO_MMAP 
			| SNDRV_PCM_INFO_INTERLEAVED 
			| SNDRV_PCM_INFO_MMAP_VALID),
#endif
	.info =			(SNDRV_PCM_INFO_INTERLEAVED),
	.formats =		USE_FORMATS,
	.rates =		USE_RATE,
	.rate_min =		USE_RATE_MIN,
	.rate_max =		USE_RATE_MAX,
	.channels_min =		USE_CHANNELS_MIN,
	.channels_max =		USE_CHANNELS_MAX,
	.buffer_bytes_max =	MAX_BUFFER_SIZE,
	.period_bytes_min =	64,
	.period_bytes_max =	MAX_BUFFER_SIZE,
	.periods_min =		USE_PERIODS_MIN,
	.periods_max =		USE_PERIODS_MAX,
	.fifo_size =		0,
};

static ts_snd_pcm_hardware_t snd_card_thinc_capture =
{
#if 0
	.info =		(SNDRV_PCM_INFO_MMAP 
			| SNDRV_PCM_INFO_INTERLEAVED 
			| SNDRV_PCM_INFO_MMAP_VALID),
#endif
	.info =			(SNDRV_PCM_INFO_INTERLEAVED),
	.formats =		USE_FORMATS,
	.rates =		USE_RATE,
	.rate_min =		USE_RATE_MIN,
	.rate_max =		USE_RATE_MAX,
	.channels_min =		USE_CHANNELS_MIN,
	.channels_max =		USE_CHANNELS_MAX,
	.buffer_bytes_max =	MAX_BUFFER_SIZE,
	.period_bytes_min =	64,
	.period_bytes_max =	MAX_BUFFER_SIZE,
	.periods_min =		USE_PERIODS_MIN,
	.periods_max =		USE_PERIODS_MAX,
	.fifo_size =		0,
};

/* This is the last function that is called when a stream is finished */
static void snd_card_thinc_runtime_free(ts_snd_pcm_runtime_t *runtime)
{
	snd_card_thinc_pcm_t *dpcm = runtime->private_data;
	struct thinc_snd_fifo *fifo;

	if (!dpcm) {
		printk(KERN_WARNING "thinc_snd: Warning -- audio application "
				"attempting to free runtime without non-null "
				"snd_card_thinc_pcm_t handle.\n");
		return;
	}

	fifo = dpcm->fifo;
#if ALSA_DEBUG
	printk(KERN_INFO "thinc_snd: snd_card_thinc_runtime_free.\n");
#endif
	kfree(dpcm);
#if THINC_STATS
	if (fifo) {
		printk(KERN_INFO "thinc_snd: total bytes written: %u\n", 
				fifo->total);
		fifo->total = 0;
	}
#endif

#if 0
	/* Create close message for tsd */
	thinc_snd_cmd_close(fifo);
#endif
}

/* This is the first function that is called when a stream is opened */
static int snd_card_thinc_playback_open(ts_snd_pcm_substream_t * substream)
{
	ts_snd_pcm_runtime_t *runtime = substream->runtime;
	snd_card_thinc_pcm_t *dpcm;
	int err;

#if ALSA_DEBUG
	printk(KERN_INFO "thinc_snd: snd_card_thinc_playback_open.\n");
#endif

	/* kcalloc was introduced in 2.6.9 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	dpcm = snd_magic_kcalloc(snd_card_thinc_pcm_t, 0, GFP_KERNEL);
	if (dpcm == NULL)
		return -ENOMEM;
	runtime->dma_area = snd_malloc_pages_fallback(MAX_BUFFER_SIZE,
		GFP_KERNEL, &runtime->dma_bytes);
	if(!runtime->dma_area) {
		snd_magic_kfree(dpcm);
		return -ENOMEM;
	}
	memset(runtime->dma_area, 0, runtime->dma_bytes);
#else
	dpcm = kcalloc(1, sizeof(*dpcm), GFP_KERNEL);
	if (!dpcm)
		return -ENOMEM;
#endif
	init_timer(&dpcm->timer);
	dpcm->timer.data = (unsigned long) dpcm;
	dpcm->timer.function = snd_card_thinc_pcm_timer_function;
	spin_lock_init(&dpcm->lock);
	dpcm->substream = substream;
	runtime->private_data = dpcm;
	runtime->private_free = snd_card_thinc_runtime_free;
	runtime->hw = snd_card_thinc_playback;
	if (substream->pcm->device & 1) {
		runtime->hw.info &= ~SNDRV_PCM_INFO_INTERLEAVED;
		runtime->hw.info |= SNDRV_PCM_INFO_NONINTERLEAVED;
	}
	if (substream->pcm->device & 2)
		runtime->hw.info &= 
			~(SNDRV_PCM_INFO_MMAP|SNDRV_PCM_INFO_MMAP_VALID);
	if ((err = add_playback_constraints(runtime)) < 0) {
		kfree(dpcm);
		return err;
	}

	return 0;
}

static int snd_card_thinc_capture_open(ts_snd_pcm_substream_t * substream)
{
	ts_snd_pcm_runtime_t *runtime = substream->runtime;
	snd_card_thinc_pcm_t *dpcm;
	int err;

	/* kcalloc was introduced in 2.6.9 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	dpcm = snd_magic_kcalloc(snd_card_thinc_pcm_t, 0, GFP_KERNEL);
	if (dpcm == NULL)
		return -ENOMEM;
	if ((runtime->dma_area = snd_malloc_pages_fallback(MAX_BUFFER_SIZE,
					GFP_KERNEL, &runtime->dma_bytes)) == NULL) 
	{
		snd_magic_kfree(dpcm);
		return -ENOMEM;
	}
	memset(runtime->dma_area, 0, runtime->dma_bytes);
#else
	dpcm = kcalloc(1, sizeof(*dpcm), GFP_KERNEL);
	if (dpcm == NULL)
		return -ENOMEM;
#endif
	init_timer(&dpcm->timer);
	dpcm->timer.data = (unsigned long) dpcm;
	dpcm->timer.function = snd_card_thinc_pcm_timer_function;
	spin_lock_init(&dpcm->lock);
	dpcm->substream = substream;
	runtime->private_data = dpcm;
	runtime->private_free = snd_card_thinc_runtime_free;
	runtime->hw = snd_card_thinc_capture;
	if (substream->pcm->device == 1) {
		runtime->hw.info &= ~SNDRV_PCM_INFO_INTERLEAVED;
		runtime->hw.info |= SNDRV_PCM_INFO_NONINTERLEAVED;
	}
	if (substream->pcm->device & 2)
		runtime->hw.info &= 
			~(SNDRV_PCM_INFO_MMAP|SNDRV_PCM_INFO_MMAP_VALID);
	if ((err = add_capture_constraints(runtime)) < 0) {
		kfree(dpcm);
		return err;
	}

	return 0;
}

static int snd_card_thinc_playback_close(ts_snd_pcm_substream_t * substream)
{
	struct thinc_snd_fifo *fifo = thinc_snd_fifo; 
#if ALSA_DEBUG
	printk(KERN_INFO "thinc_snd: snd_card_thinc_playback_close.\n");
#endif
	/* Create close message for tsd */
	thinc_snd_cmd_close(fifo, T_MODE_PLAYBACK);

	/* now reset the pointers */
	memset(fifo->buf, 0, T_SND_BUFSIZE);
	fifo->num_ctrl_cmds = fifo->num_play_cmds = 0;
	fifo->snd_in = T_SND_SOUND_BUFSTART;
	fifo->play_last = fifo->play_in = T_SND_CMD_BUFSTART;

	return 0;
}

static int snd_card_thinc_capture_close(ts_snd_pcm_substream_t * substream)
{
	struct thinc_snd_fifo *fifo = thinc_snd_fifo; 
	unsigned long flags;

	thinc_snd_cmd_close(fifo, T_MODE_CAPTURE);

	spin_lock_irqsave(&fifo->buf_lock, flags);

	/* now reset the pointers */
	memset(fifo->capbuf, 0, T_SND_BUFSIZE);
	/* so we ignore any new data from the driver */
	fifo->capbuf_avail = -1;
	fifo->capbuf_head = fifo->capbuf;

	spin_unlock_irqrestore(&fifo->buf_lock, flags);

	return 0;
}

static ts_snd_pcm_ops_t snd_card_thinc_playback_ops = {
	.open =			snd_card_thinc_playback_open,
	.close =		snd_card_thinc_playback_close,
	.ioctl =		snd_pcm_lib_ioctl,
	.prepare =		snd_card_thinc_playback_prepare,
	.trigger =		snd_card_thinc_playback_trigger,
	.pointer =		snd_card_thinc_playback_pointer,
	.copy = 		snd_card_thinc_playback_copy,
	.silence = 		snd_card_thinc_playback_silence,
};

static ts_snd_pcm_ops_t snd_card_thinc_capture_ops = {
	.open =			snd_card_thinc_capture_open,
	.close =		snd_card_thinc_capture_close,
	.ioctl =		snd_pcm_lib_ioctl,
	.prepare =		snd_card_thinc_capture_prepare,
	.trigger =		snd_card_thinc_capture_trigger,
	.pointer =		snd_card_thinc_capture_pointer,
	.copy =                 snd_card_thinc_capture_copy,
	.silence = 		snd_card_thinc_capture_silence,
};

/* fops for /dev/tsnd */
struct file_operations thinc_snd_fops = {
	.ioctl = thinc_snd_ioctl,
	.mmap = thinc_snd_mmap,
	.open = thinc_snd_open,
	.read = thinc_snd_read,
	.write = thinc_snd_write,
	.release = thinc_snd_release,
	.poll = thinc_snd_poll,
};

int thinc_snd_ioctl(struct inode *inode, struct file *filp, unsigned int cmd,
		unsigned long arg)
{
	struct thinc_snd_cmd_node *node;
	struct thinc_snd_fifo *fifo = filp->private_data;

	switch(cmd) {
		case T_SND_IOC_PLAY_CMD_OFFSET:
			return fifo->play_last;
		case T_SND_IOC_NUM_PLAY_CMDS:
			return fifo->num_play_cmds;
		case T_SND_IOC_DEL_PLAY_CMD:
			fifo_del_play_cmds(fifo);
			return 0;
		case T_SND_IOC_RESET_PLAY_CMDS:
			fifo_reset_play_cmds(fifo);
			return 0;
		case T_SND_IOC_OPEN_CAPTURE_STREAM:
			/* XXX: Huh? */
			return 0;
		case T_SND_IOC_OPEN_CTRL_CMD:
			node = thinc_snd_make_cmd_node(fifo, 
				T_SND_CMD_PLAYBACK_OPEN);
			if (node == NULL) {
				return -ENOMEM;
			}
			memcpy(&node->cmd.type.open, &fifo->open_params,
					sizeof(struct thinc_snd_cmd_open));
			return 0;
	}

	/* should never get here */
	return -1;
}

int thinc_snd_open(struct inode *inode, struct file *filp)
{
	filp->f_op = &thinc_snd_fops;
	/* :TODO: how can we have user AND session-specific fifos? */
	filp->private_data = thinc_snd_fifo;

	if (!thinc_snd_fifo) return -EINVAL;

	if (thinc_snd_fifo->connected == 0) {
		fifo_reset_play_cmds(thinc_snd_fifo);
	}

	thinc_snd_fifo->connected++;

	try_module_get(THIS_MODULE);

	return 0;
}

/* This writes a thinc_snd_cmd_* struct to tsd */
ssize_t thinc_snd_read(struct file *filp, char *buf, size_t count,
	loff_t *offset)
{   
	ssize_t retval;
	struct thinc_snd_cmd_node *node;
	struct thinc_snd_fifo *fifo = filp->private_data;

	/* If no commands, return nothing */
	if (list_empty(&fifo->ctrl_commands)) return 0;

	retval = sizeof(struct thinc_snd_cmd);
	node = list_entry(fifo->ctrl_commands.next, struct thinc_snd_cmd_node,
			list);

	if (copy_to_user(buf, &node->cmd, retval) > 0) {
		/* What do we fricking do? copy_to_user could not copy all its 
		 * bytes to tsd. */
		return -1;
	}

	/* Now, remove the node and free it */
	list_del(&node->list);
	kfree(node);

	return retval;
}       

/* this reads how much data was written to the mmap'ed capture buffer so we
 * know how much is available */ 
ssize_t thinc_snd_write(struct file *filp, const char __user *buf, 
	size_t count, loff_t *offset)
{
#if 0	/* CRAP */
	int i, j, k, block_size, add;
	struct thinc_snd_fifo *fifo = filp->private_data;
	ssize_t retval = count;

	if(count == 0) 
		return 0;

	/* copy contents of buf into fifo->capbuf */
	block_size = T_SND_CAPTURE_BLOCKSIZE;
	add = (count + T_SND_CAPTURE_BLOCKSIZE - 1) / T_SND_CAPTURE_BLOCKSIZE;

	for(i=0; i<add; i++) {
		j = ((fifo->cb_end + i) % fifo->cb_length) * 
			T_SND_CAPTURE_BLOCKSIZE;
		k = i * T_SND_CAPTURE_BLOCKSIZE;
		retval -= copy_from_user(fifo->capbuf + j, buf + k, 
			block_size);
	}

	if(fifo->cb_numblocks + add > fifo->cb_length) {
		fifo->cb_numblocks = fifo->cb_length;
	} else {
		fifo->cb_numblocks = fifo->cb_numblocks + add;
	}
	fifo->cb_begin = (fifo->cb_length + fifo->cb_end + add - 
		fifo->cb_numblocks) % fifo->cb_length;
	fifo->cb_end = (fifo->cb_end + add) % fifo->cb_length;

	return retval;
#else
	struct thinc_snd_fifo *fifo = filp->private_data;
	int how_much;
	unsigned long flags;
#if 0
unsigned long us;

gettime_us(us);
#endif

	/* be safe */
	if (count != sizeof(int)) {
		printk(KERN_WARNING "thinc_snd: I was expecting an integer, "
			"you sent me %d bytes. what's up?", count);
		return -EINVAL;
	}
	if(get_user(how_much, (int __user*) buf)) 
		return -EFAULT;

	spin_lock_irqsave(&fifo->buf_lock, flags);
	if(fifo->capbuf_avail == -1) {
		/* we're not capturing data, just ignore */
#if THINC_DEBUG
		printk(KERN_INFO "thinc_snd: client sent %d bytes, but we're "
			"not capturing anymore, ignoring\n", how_much);
#endif
		spin_unlock_irqrestore(&fifo->buf_lock, flags);
		return count;
	}

	/* keep track of how much data we have available */
	fifo->capbuf_avail += how_much;
	if(fifo->capbuf_avail > T_SND_BUFSIZE)
		fifo->capbuf_avail = T_SND_BUFSIZE;
	/* advance the head, but careful as it may wrap around */
	fifo->capbuf_head += how_much;
	if(fifo->capbuf_head > fifo->capbuf_tail) {
		/* wrap around */
		fifo->capbuf_head = fifo->capbuf + 
			(fifo->capbuf_head - fifo->capbuf_tail);
	}

#if THINC_DEBUG
	printk(KERN_INFO "thinc_snd: client sent %d bytes, now have %d "
		"available\n", how_much, fifo->capbuf_avail);
#endif

#if 0
/* XXX DEBUG */
printk(KERN_INFO "thinc_snd: [%lu] write h: %d g: %u\n", us, fifo->capbuf_avail, how_much);
#endif

	spin_unlock_irqrestore(&fifo->buf_lock, flags);

	return count;
#endif
}

int thinc_snd_release(struct inode *inode, struct file *filp)
{
	struct thinc_snd_fifo *fifo = filp->private_data;

#if THINC_DEBUG
	printk(KERN_INFO "thinc_snd: tsd releasing driver. cleared command queue.\n");
#endif

	/* clear the command list. */
	if (fifo->connected == 0) {
		thinc_snd_free_cmds(fifo);
	}

	fifo->connected--;

	/* decrement module usage count */
	module_put(THIS_MODULE);

	return 0;
}

unsigned int thinc_snd_poll(struct file *filp, poll_table *wait)
{
	unsigned int mask = 0;
	struct thinc_snd_fifo *fifo = filp->private_data;

	poll_wait(filp, &fifo->waitq_in, wait);

	/* return if there a command to process. */
	if (fifo->num_play_cmds > 0 || !list_empty(&fifo->ctrl_commands)) {
		mask |= POLLIN | POLLRDNORM;
	}

	return mask;
}

/*************************** THINC FILE OPERATIONS ****************************/
int thinc_snd_mmap(struct file *filp, struct vm_area_struct *vma)
{
	unsigned long page, pos;
	unsigned long start = (unsigned long)vma->vm_start;
	unsigned long size = (unsigned long)(vma->vm_end-vma->vm_start);
	struct thinc_snd_fifo *fifo = filp->private_data;

	/* If userspace tries to mmap beyond end of our buffer, fail */
	if (size > T_SND_BUFSIZE) return -EINVAL;

	/* check which buffer the daemon wants.for now pretty raw :) */
	if(vma->vm_pgoff == 0) {
		/* wants playback buffer */
		pos = (unsigned long) fifo->buf;
	} else {
		pos = (unsigned long) fifo->capbuf;
	}

#if ( (LINUX_VERSION_CODE) < (KERNEL_VERSION(2,6,13)) )
	/* Loop through all the physical pages in the buffer */
	/* NOTE: This won't work for vmalloc()d memory ! */
	while (size > 0) {
		/* Remap a single physical page to the process's vma */
		page = virt_to_phys((void *)pos);
		/* Fifth argument is the protection of the map.  */
		if (io_remap_page_range(vma, start, page, PAGE_SIZE, 
					PAGE_SHARED)) {
			return -EINVAL;
		}
		start += PAGE_SIZE;
		pos += PAGE_SIZE;
		size -= PAGE_SIZE;
	}
#else
	/* Remap a single physical page to the process's vma */
	page = virt_to_phys((void *)pos);
	/* Fifth argument is the protection of the map.  */
	if (io_remap_pfn_range(vma, start, page >> PAGE_SHIFT, size, 
		PAGE_SHARED)) {
		return -EINVAL;
	}
#endif
	return 0;
}

static int __init snd_card_thinc_pcm(snd_card_thinc_t *thinc, int device, 
		int substreams)
{
	ts_snd_pcm_t *pcm;
	int err;

#if ALSA_DEBUG
	printk(KERN_INFO "thinc_snd: snd_card_thinc_pcm\n");
#endif

	if ((err = snd_pcm_new(thinc->card, "THINC PCM", device, substreams, 
					substreams, &pcm)) < 0) 
	{
		return err;
	}
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, 
			&snd_card_thinc_playback_ops);
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, 
			&snd_card_thinc_capture_ops);
	pcm->private_data = thinc;
	pcm->info_flags = 0;
	strcpy(pcm->name, "THINC PCM");
	/* I'm not sure why this needs such a large chunk of memory. */
	snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_CONTINUOUS,
			snd_dma_continuous_data(GFP_KERNEL), 0, 64*1024);
	return 0;
}

static int snd_card_thinc_playback_silence(ts_snd_pcm_substream_t *substream,
		int channel, snd_pcm_uframes_t pos, snd_pcm_uframes_t count)
{   
#if ALSA_DEBUG
	printk(KERN_INFO "thinc_snd: snd_thinc_playback_silence called.\n");
#endif
	return 0;
}

#if 0
/* NO-OP used when no daemon has opened this driver. */
static int snd_card_thinc_playback_prepare_noop(ts_snd_pcm_substream_t *substream)
{
	return -EINVAL;
}

static int snd_card_thinc_playback_copy_noop(ts_snd_pcm_substream_t *substream,
		int channel, snd_pcm_uframes_t pos, void __user *src, 
		snd_pcm_uframes_t count)
{
	/* In case tsd is killed and some sound program is still running. The
	 * sound app will still play, but it will complain loudly. If tsd dies
	 * while the program is running and is restarted, sound probably won't
	 * restart until another sound is played again. */
	return -EINVAL;
}
#endif

static int snd_card_thinc_capture_silence(ts_snd_pcm_substream_t *substream,
		int channel, snd_pcm_uframes_t pos, snd_pcm_uframes_t count)
{
#if ALSA_DEBUG
	printk(KERN_INFO "thinc_snd: snd_thinc_capture_silence called.\n");
#endif
	return 0;
}

/* the main capture function */
#if 0	/* CRAP */
/* FIXME: Probably inefficient too? */
static int snd_card_thinc_capture_copy(ts_snd_pcm_substream_t *substream,
	int channel, snd_pcm_uframes_t pos, void __user *src, 
	snd_pcm_uframes_t count)
{
	int i, j, k, block_size, sub, diff, sent;
	ts_snd_pcm_runtime_t *runtime = substream->runtime;
	snd_card_thinc_pcm_t *dpcm = runtime->private_data;
	struct thinc_snd_fifo *fifo = dpcm->fifo;
	int framesize = (runtime->sample_bits * runtime->channels) / 8;

#if ALSA_DEBUG
	printk(KERN_INFO "thinc_snd: %s copying %lu frames\n", __func__,
		count);
#endif

	block_size = T_SND_CAPTURE_BLOCKSIZE;
	sub = ((count*framesize) + T_SND_CAPTURE_BLOCKSIZE - 1) / T_SND_CAPTURE_BLOCKSIZE;

	/* user is requesting more data than fits in the capture buffer */
	if(sub > fifo->cb_length) {
		return -EINVAL;
	}

	/* user is requesting more data than is currently available */
	if(sub > fifo->cb_numblocks) {
#if 0
		sub = fifo->cb_numblocks;    
#else
		return -EWOULDBLOCK;
#endif
	}

	sent = 0;
	for(i=0; i<sub; i++) {
		j = (fifo->cb_begin % fifo->cb_length)*T_SND_CAPTURE_BLOCKSIZE;
		k = i * T_SND_CAPTURE_BLOCKSIZE;
		if(count*4 < block_size) {
			diff = copy_to_user(src + k, fifo->capbuf + j, 
				count*framesize);
		} else {
			diff = copy_to_user(src + k, fifo->capbuf + j, 
				block_size);
		}
		sent += (block_size - diff);
		fifo->cb_numblocks--;
		fifo->cb_begin = (fifo->cb_begin + 1) % fifo->cb_length;
	} 
	return 0;
}
#else
static int snd_card_thinc_capture_copy(ts_snd_pcm_substream_t *substream,
	int channel, snd_pcm_uframes_t pos, void __user *src, 
	snd_pcm_uframes_t count)
{
	ts_snd_pcm_runtime_t *runtime = substream->runtime;
	snd_card_thinc_pcm_t *dpcm = runtime->private_data;
	struct thinc_snd_fifo *fifo = dpcm->fifo;
	unsigned long flags;

#if 0
	/* We do not use position, instead we keep track of position 
	 * ourselves. Note that this position is given to us by ALSA
	 * based on the buffer size specified by MAX_BUFFER_SIZE. We 
	 * could potentially optimize our fifo queue handling by using
	 * this. However, AFAICT MAX_BUFFER_SIZE is capped at 64KB, 
	 * and this size affects synchronization, so care must be taken
	 * if we muck with this.
	 */
	pos = frames_to_bytes(runtime, pos);
#endif
	count = frames_to_bytes(runtime, count);
	if (count > T_SND_BUFSIZE) {
		/* something is seriously wrong here. the user is trying to 
		 * read an absurd amount of data. warn! */
		printk(KERN_WARNING "thinc_snd: User attempting to read %lu "
			"bytes of sound data when there is only %d of "
			"buffer space.\n", count, T_SND_SOUND_BUFSIZE);
		return -EFAULT;
	}

	/* if user asking for too much, be nice and let her know */
	/* FIXME: Problem is, stupid skype dies if we do this :( so just
	 * return crap :) */
	spin_lock_irqsave(&fifo->buf_lock, flags);
	if(count > fifo->capbuf_avail) {
#if THINC_DEBUG
		printk(KERN_INFO "thinc_snd: user wants %ld bytes, we only "
			"have %d, sorry\n", count, fifo->capbuf_avail);
#endif
#if 0
{unsigned long us;
gettime_us(us);
printk(KERN_INFO "thinc_snd: [%lu] read h: %d w: %lu not enough\n", us, fifo->capbuf_avail, count);
}
#endif
#if 0
		return -EWOULDBLOCK;
#else
		spin_unlock_irqrestore(&fifo->buf_lock, flags);
		return 0;
#endif
	}
	spin_unlock_irqrestore(&fifo->buf_lock, flags);

	if (fifo_get_sound(fifo, src, count) == -EFAULT) {
		return -EFAULT;
	}
	return 0;
}
#endif

/* Used when the driver has been mmap'ed */
static int snd_card_thinc_playback_copy(ts_snd_pcm_substream_t *substream,
	int channel, snd_pcm_uframes_t pos, void __user *src, 
	snd_pcm_uframes_t count)
{
	unsigned int stamp;
	ts_snd_pcm_runtime_t *runtime = substream->runtime;
	snd_card_thinc_pcm_t *dpcm = runtime->private_data;
	struct thinc_snd_fifo *fifo = dpcm->fifo;

#if 0
	/* We do not use position, instead we keep track of position 
	 * ourselves. Note that this position is given to us by ALSA
	 * based on the buffer size specified by MAX_BUFFER_SIZE. We 
	 * could potentially optimize our fifo queue handling by using
	 * this. However, AFAICT MAX_BUFFER_SIZE is capped at 64KB, 
	 * and this size affects synchronization, so care must be taken
	 * if we muck with this.
	 */
	pos = frames_to_bytes(runtime, pos);
#endif
	count = frames_to_bytes(runtime, count);
	if (count > T_SND_SOUND_BUFSIZE) {
		/* something is seriously wrong here. the user is trying to 
		 * write an absurd amount of data. warn! */
		printk(KERN_WARNING "thinc_snd: User attempting to write %lu "
			"bytes of sound data when there is only %d of "
			"buffer space.\n", count, T_SND_SOUND_BUFSIZE);
		return -EFAULT;
	}

	/* get the stamp */
	gettime_us(stamp);
#if THINC_STATS
	printk(KERN_INFO "thinc_snd: sound timestamp is %u\n", stamp);
#endif
	if ( fifo_put_sound(fifo, src, stamp, count) == -EFAULT) {
		return -EFAULT;
	}

#if THINC_STATS
	printk(KERN_INFO "thinc_snd: wrote %u bytes\n", (unsigned int)count);
	fifo->total += count;
#endif
	return 0;
}

#define THINC_VOLUME(xname, xindex, addr) \
{ .iface = SNDRV_CTL_ELEM_IFACE_MIXER, .name = xname, .index = xindex, \
	.info = snd_thinc_volume_info, \
	.get = snd_thinc_volume_get, .put = snd_thinc_volume_put, \
	.private_value = addr }

static int snd_thinc_volume_info(ts_snd_kcontrol_t * kcontrol, 
	ts_snd_ctl_elem_info_t * uinfo)
{
#if ALSA_DEBUG
	printk(KERN_INFO "thinc_snd: snd_thinc_volume_info called.\n");
#endif
	uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	uinfo->count = 2;
	uinfo->value.integer.min = -50;
	uinfo->value.integer.max = 100;
	return 0;
}

static int snd_thinc_volume_get(ts_snd_kcontrol_t * kcontrol, 
	ts_snd_ctl_elem_value_t * ucontrol)
{
	snd_card_thinc_t *thinc = snd_kcontrol_chip(kcontrol);
	unsigned long flags;
	int addr = kcontrol->private_value;

#if ALSA_DEBUG
	printk(KERN_INFO "thinc_snd: snd_thinc_volume_get called.\n");
#endif
	spin_lock_irqsave(&thinc->mixer_lock, flags);
	ucontrol->value.integer.value[0] = thinc->mixer_volume[addr][0];
	ucontrol->value.integer.value[1] = thinc->mixer_volume[addr][1];
	spin_unlock_irqrestore(&thinc->mixer_lock, flags);
	return 0;
}

static int snd_thinc_volume_put(ts_snd_kcontrol_t * kcontrol, 
	ts_snd_ctl_elem_value_t * ucontrol)
{
	snd_card_thinc_t *thinc = snd_kcontrol_chip(kcontrol);
	unsigned long flags;
	int change, addr = kcontrol->private_value;
	int left, right;

#if ALSA_DEBUG
	printk(KERN_INFO "thinc_snd: snd_thinc_volume_put called.\n");
#endif
	left = ucontrol->value.integer.value[0];
	if (left < -50)
		left = -50;
	if (left > 100)
		left = 100;
	right = ucontrol->value.integer.value[1];
	if (right < -50)
		right = -50;
	if (right > 100)
		right = 100;
	spin_lock_irqsave(&thinc->mixer_lock, flags);
	change = thinc->mixer_volume[addr][0] != left ||
		thinc->mixer_volume[addr][1] != right;
	thinc->mixer_volume[addr][0] = left;
	thinc->mixer_volume[addr][1] = right;
	spin_unlock_irqrestore(&thinc->mixer_lock, flags);
	return change;
}

#define THINC_CAPSRC(xname, xindex, addr) \
{ .iface = SNDRV_CTL_ELEM_IFACE_MIXER, .name = xname, .index = xindex, \
	.info = snd_thinc_capsrc_info, \
	.get = snd_thinc_capsrc_get, .put = snd_thinc_capsrc_put, \
	.private_value = addr }

static int snd_thinc_capsrc_info(ts_snd_kcontrol_t * kcontrol, 
	ts_snd_ctl_elem_info_t * uinfo)
{
#if ALSA_DEBUG
	printk(KERN_INFO "thinc_snd: calling 'snd_thinc_capsrc_info'\n");
#endif
	uinfo->type = SNDRV_CTL_ELEM_TYPE_BOOLEAN;
	uinfo->count = 2;
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = 1;
	return 0;
}

static int snd_thinc_capsrc_get(ts_snd_kcontrol_t * kcontrol, 
	ts_snd_ctl_elem_value_t * ucontrol)
{
	snd_card_thinc_t *thinc = snd_kcontrol_chip(kcontrol);
	unsigned long flags;
	int addr = kcontrol->private_value;

	spin_lock_irqsave(&thinc->mixer_lock, flags);
	ucontrol->value.integer.value[0] = thinc->capture_source[addr][0];
	ucontrol->value.integer.value[1] = thinc->capture_source[addr][1];
	spin_unlock_irqrestore(&thinc->mixer_lock, flags);
	return 0;
}

static int snd_thinc_capsrc_put(ts_snd_kcontrol_t * kcontrol, 
	ts_snd_ctl_elem_value_t * ucontrol)
{
	snd_card_thinc_t *thinc = snd_kcontrol_chip(kcontrol);
	unsigned long flags;
	int change, addr = kcontrol->private_value;
	int left, right;

	left = ucontrol->value.integer.value[0] & 1;
	right = ucontrol->value.integer.value[1] & 1;
	spin_lock_irqsave(&thinc->mixer_lock, flags);
	change = thinc->capture_source[addr][0] != left &&
		thinc->capture_source[addr][1] != right;
	thinc->capture_source[addr][0] = left;
	thinc->capture_source[addr][1] = right;
	spin_unlock_irqrestore(&thinc->mixer_lock, flags);
	return change;
}

static ts_snd_kcontrol_new_t snd_thinc_controls[] = {
	THINC_VOLUME("Master Volume", 0, MIXER_ADDR_MASTER),
	THINC_CAPSRC("Master Capture Switch", 0, MIXER_ADDR_MASTER),
	THINC_VOLUME("Synth Volume", 0, MIXER_ADDR_SYNTH),
	THINC_CAPSRC("Synth Capture Switch", 0, MIXER_ADDR_MASTER),
	THINC_VOLUME("Line Volume", 0, MIXER_ADDR_LINE),
	THINC_CAPSRC("Line Capture Switch", 0, MIXER_ADDR_MASTER),
	THINC_VOLUME("Mic Volume", 0, MIXER_ADDR_MIC),
	THINC_CAPSRC("Mic Capture Switch", 0, MIXER_ADDR_MASTER),
	THINC_VOLUME("CD Volume", 0, MIXER_ADDR_CD),
	THINC_CAPSRC("CD Capture Switch", 0, MIXER_ADDR_MASTER)
};

int __init snd_card_thinc_new_mixer(snd_card_thinc_t * thinc)
{
	ts_snd_card_t *card = thinc->card;
	unsigned int idx;
	int err;

#if ALSA_DEBUG
	printk(KERN_INFO "thinc_snd: calling 'snd_card_thinc_new_mixer'\n");
#endif
	snd_assert(thinc != NULL, return -EINVAL);
	spin_lock_init(&thinc->mixer_lock);
	strcpy(card->mixername, "THINC Mixer");

	for (idx = 0; idx < ARRAY_SIZE(snd_thinc_controls); idx++) {
		if ((err = snd_ctl_add(card, 
			snd_ctl_new1(&snd_thinc_controls[idx], thinc))) < 0)
		{
			return err;
		}
	}
	return 0;
}

static int __init snd_card_thinc_probe(int dev)
{
	ts_snd_card_t *card;
	struct snd_card_thinc *thinc;
	int idx, err;

#if ALSA_DEBUG
	printk(KERN_INFO "thinc_snd: calling 'snd_card_thinc_probe'\n");
#endif
	if (!enable[dev])
		return -ENODEV;
	card = snd_card_new(index[dev], id[dev], THIS_MODULE,
			sizeof(struct snd_card_thinc));
	if (card == NULL)
		return -ENOMEM;
	thinc = (struct snd_card_thinc *)card->private_data;
	thinc->card = card;
	for (idx = 0; idx < MAX_PCM_DEVICES && idx < pcm_devs[dev]; idx++) {
		if (pcm_substreams[dev] < 1)
			pcm_substreams[dev] = 1;
		if (pcm_substreams[dev] > MAX_PCM_SUBSTREAMS)
			pcm_substreams[dev] = MAX_PCM_SUBSTREAMS;
		if ((err = snd_card_thinc_pcm(thinc, idx, 
						pcm_substreams[dev])) < 0) 
		{
			goto __nodev;
		}
	}
	if ((err = snd_card_thinc_new_mixer(thinc)) < 0)
		goto __nodev;
	strcpy(card->driver, "THINC");
	strcpy(card->shortname, "THINC");
	sprintf(card->longname, "THINC %i", dev + 1);
	if ((err = snd_card_register(card)) == 0) {
		snd_thinc_cards[dev] = card;
		return 0;
	}
__nodev:
	snd_card_free(card);
	return err;
}

static void __exit alsa_card_thinc_exit(void)
{
	int idx;

#if ALSA_DEBUG
	printk(KERN_INFO "thinc_snd: calling 'alsa_card_thinc_exit'\n");
#endif

	for (idx = 0; idx < SNDRV_CARDS; idx++) {
		snd_card_free(snd_thinc_cards[idx]);
	}

	/* free the buffer separately. */
	if (thinc_snd_fifo->buf) {
		snd_free_pages(thinc_snd_fifo->buf, T_SND_BUFSIZE);
	}
	if (thinc_snd_fifo->capbuf) {
		snd_free_pages(thinc_snd_fifo->capbuf, T_SND_BUFSIZE);
	}

	/* free fifo! */
	if (thinc_snd_fifo) {
		kfree(thinc_snd_fifo);
	}

	/* cleanup_module is never called if registering failed */
	unregister_chrdev(T_SND_MAJOR, T_SND_NAME);

#if ALSA_DEBUG
	printk(KERN_INFO "thinc_snd: bye bye\n");
#endif
}

static int __init alsa_card_thinc_init(void)
{
	int dev, cards;
	int retval;

	for (dev = cards = 0; dev < SNDRV_CARDS && enable[dev]; dev++) {
		if (snd_card_thinc_probe(dev) < 0) {
#ifdef MODULE
			printk(KERN_ERR "THINC soundcard #%i not found or "
				"device busy\n", dev + 1);
#endif
			break;
		}
		cards++;
	}
	if (!cards) {
#ifdef MODULE
		printk(KERN_ERR "THINC soundcard not found or device busy\n");
#endif
		return -ENODEV;
	}

#if THINC_DEBUG
	printk(KERN_INFO "thinc_snd: initting\n");
#endif

	/*
	 * Register major, and accept a dynamic number. This is the first thing
	 * to do, in order to avoid releasing other module's fops while
	 * cleaning up.
	 */
	retval = register_chrdev(T_SND_MAJOR, T_SND_NAME,
			&thinc_snd_fops);

#ifdef MODULE
	if (retval < 0) {
		printk(KERN_WARNING "thinc_snd: can't get major %d\n", 
			T_SND_MAJOR);
		return retval;
	}
#endif

	/* init the fifo queue */
	if ( IS_ERR(thinc_snd_fifo = fifo_init())) {
		printk(KERN_INFO "thinc_snd: failed fifo_init -- no memory available\n");
		alsa_card_thinc_exit();
		return PTR_ERR(thinc_snd_fifo);
	}

#if THINC_DEBUG
	printk("-----------------------------------------------------------\n");
	printk("thinc_snd driver initted\n");
	printk("-----------------------------------------------------------\n");
#endif
	return 0;
}

/* debug */
#if ALSA_DEBUG
static void debug_substream(ts_snd_pcm_substream_t *substream)
{
	printk("----------------------------------------------------------\n");
	printk("thinc_snd: substream attributes\n");
	printk("----------------------------------------------------------\n");
	printk("snd_card->number: '%d'\n", substream->pcm->card->number);
	printk("snd_card->id string: '%s'\n", substream->pcm->card->id);
	printk("snd_card->driver name: '%s'\n", substream->pcm->card->driver);
	printk("snd_card->short name: '%s'\n", substream->pcm->card->shortname);
	printk("snd_card->long name: '%s'\n", substream->pcm->card->longname);
	printk("snd_card->mixer name: '%s'\n", substream->pcm->card->mixername);
	printk("snd_card->components: '%s'\n", substream->pcm->card->components);
	printk("snd_pcm->device number: '%u'\n", substream->pcm->device);
	printk("snd_pcm->dev class: '%u'\n", substream->pcm->dev_class);
	printk("snd_pcm->dev subclass: '%u'\n", substream->pcm->dev_subclass);
	printk("snd_pcm->id: '%s'\n", substream->pcm->id);
	printk("snd_pcm->name: '%s'\n", substream->pcm->name);
	printk("snd_pcm_str->stream (direction): '%d'\n", substream->pstr->stream);
	printk("snd_pcm_substream->number: '%d'\n", substream->number);
	printk("snd_pcm_substream->name: '%s'\n", substream->name);
	printk("snd_pcm_substream->stream (direction): '%s'\n", substream->name);
	printk("snd_pcm_substream->ring buffer size limit: '%d'\n", 
			substream->buffer_bytes_max);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,9)
	printk("snd_pcm_substream->dma_buf_id: '%u'\n", substream->dma_buf_id);
#endif
	printk("snd_pcm_substream->dma_max: '%d'\n", substream->dma_max);
	printk("snd_pcm_substream->dma_addr: '%d'\n", 
			substream->runtime->dma_addr);
	if (substream->runtime->dma_buffer_p == NULL) {
		printk("snd_pcm_substream->runtime->dma_buffer_p WAS NULL\n");
	} else {
		printk("snd_pcm_substream->runtime->dma_buffer_p WAS NOT NULL\n");
	}
	printk(KERN_INFO "runtime->rate: %u\n", substream->runtime->rate);
	printk(KERN_INFO "runtime->channels: %u\n", 
			substream->runtime->channels);
	printk(KERN_INFO "runtime->access mode: %u\n", 
			substream->runtime->access);
	printk(KERN_INFO "runtime->frame_bits: %u\n", 
			substream->runtime->frame_bits);
	printk(KERN_INFO "runtime->sample_bits: %u\n", 
			substream->runtime->sample_bits);
	printk(KERN_INFO "is big endian: %d\n",
			snd_pcm_format_big_endian(substream->runtime->format));
}
#endif

module_init(alsa_card_thinc_init)
module_exit(alsa_card_thinc_exit)
